@@ -60,6 +60,210 @@ pyg_type_wrapper_dealloc(PyGTypeWrapper *self)
     PyObject_DEL(self);
 }
 
+static PyObject *
+_wrap_g_type_wrapper__get_pytype(PyGTypeWrapper *self, void *closure)
+{
+    PyObject *py_type;
+    
+    py_type = g_type_get_qdata(self->type, pygobject_class_key);
+    if (!py_type)
+      py_type = Py_None;
+
+    Py_INCREF(py_type);
+    return py_type;
+}
+
+static PyObject *
+_wrap_g_type_wrapper__get_name(PyGTypeWrapper *self, void *closure)
+{
+   const char *name = g_type_name(self->type);
+   return PyString_FromString(g_strdup(name ? name : "invalid"));
+}
+
+static PyObject *
+_wrap_g_type_wrapper__get_parent(PyGTypeWrapper *self, void *closure)
+{
+   return pyg_type_wrapper_new(g_type_parent(self->type));
+}
+
+static PyObject *
+_wrap_g_type_wrapper__get_fundamental(PyGTypeWrapper *self, void *closure)
+{
+   return pyg_type_wrapper_new(g_type_fundamental(self->type));
+}
+
+static PyObject *
+_wrap_g_type_wrapper__get_children(PyGTypeWrapper *self, void *closure)
+{
+  int n_children, i;
+  GType *children;
+  PyObject *retval;
+    
+  children = g_type_children(self->type, &n_children);
+
+  retval = PyList_New(n_children);
+  for (i = 0; i < n_children; i++)
+      PyList_SetItem(retval, i, pyg_type_wrapper_new(children[i]));
+  g_free(children);
+  
+  return retval;
+}
+
+static PyObject *
+_wrap_g_type_wrapper__get_interfaces(PyGTypeWrapper *self, void *closure)
+{
+  int n_interfaces, i;
+  GType *interfaces;
+  PyObject *retval;
+    
+  interfaces = g_type_interfaces(self->type, &n_interfaces);
+
+  retval = PyList_New(n_interfaces);
+  for (i = 0; i < n_interfaces; i++)
+      PyList_SetItem(retval, i, pyg_type_wrapper_new(interfaces[i]));
+  g_free(interfaces);
+  
+  return retval;
+}
+
+static PyObject *
+_wrap_g_type_wrapper__get_depth(PyGTypeWrapper *self, void *closure)
+{
+  return PyInt_FromLong(g_type_depth(self->type));
+}
+
+static PyGetSetDef _PyGTypeWrapper_getsets[] = {
+    { "pytype", (getter)_wrap_g_type_wrapper__get_pytype, (setter)0 },
+    { "name",  (getter)_wrap_g_type_wrapper__get_name, (setter)0 },
+    { "fundamental",  (getter)_wrap_g_type_wrapper__get_fundamental, (setter)0 },
+    { "parent",  (getter)_wrap_g_type_wrapper__get_parent, (setter)0 },
+    { "children",  (getter)_wrap_g_type_wrapper__get_children, (setter)0 },
+    { "interfaces",  (getter)_wrap_g_type_wrapper__get_interfaces, (setter)0 },
+    { "depth",  (getter)_wrap_g_type_wrapper__get_depth, (setter)0 },
+    { NULL, (getter)0, (setter)0 }
+};
+
+static PyObject*
+_wrap_g_type_is_interface(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_INTERFACE(self->type));
+}
+
+static PyObject*
+_wrap_g_type_is_classed(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_CLASSED(self->type));
+}
+
+static PyObject*
+_wrap_g_type_is_instantiatable(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_INSTANTIATABLE(self->type));
+}
+
+static PyObject*
+_wrap_g_type_is_derivable(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_DERIVABLE(self->type));
+}
+
+static PyObject*
+_wrap_g_type_is_deep_derivable(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_DEEP_DERIVABLE(self->type));
+}
+
+static PyObject*
+_wrap_g_type_is_abstract(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_ABSTRACT(self->type));
+}
+
+static PyObject*
+_wrap_g_type_is_value_abstract(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_VALUE_ABSTRACT(self->type));
+}
+
+static PyObject*
+_wrap_g_type_is_value_type(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_IS_VALUE_TYPE(self->type));
+}
+
+static PyObject*
+_wrap_g_type_has_value_table(PyGTypeWrapper *self)
+{
+    return PyBool_FromLong(G_TYPE_HAS_VALUE_TABLE(self->type));
+}
+
+static PyObject*
+_wrap_g_type_from_name(PyGTypeWrapper *_, PyObject *args)
+{
+    char *type_name;
+    GType type;
+    
+    if (!PyArg_ParseTuple(args, "s:GType.from_name", &type_name))
+	return NULL;
+
+    type = g_type_from_name(type_name);
+    if (type == 0) {
+	PyErr_SetString(PyExc_RuntimeError, "unknown type name");
+	return NULL;
+    }
+
+    return pyg_type_wrapper_new(type);
+}
+
+static PyObject*
+_wrap_g_type_is_a(PyGTypeWrapper *self, PyObject *args)
+{
+    PyObject *gparent;
+    GType parent;
+
+    if (!PyArg_ParseTuple(args, "O:GType.is_a", &gparent))
+	return NULL;
+    else if ((parent = pyg_type_from_object(gparent)) == 0)
+	return NULL;
+    
+    return PyBool_FromLong(g_type_is_a(self->type, parent));
+}
+
+static PyMethodDef _PyGTypeWrapper_methods[] = {
+    { "is_interface", (PyCFunction)_wrap_g_type_is_interface, METH_NOARGS },
+    { "is_classed", (PyCFunction)_wrap_g_type_is_classed, METH_NOARGS },
+    { "is_instantiatable", (PyCFunction)_wrap_g_type_is_instantiatable, METH_NOARGS },
+    { "is_derivable", (PyCFunction)_wrap_g_type_is_derivable, METH_NOARGS },
+    { "is_deep_derivable", (PyCFunction)_wrap_g_type_is_deep_derivable, METH_NOARGS },
+    { "is_abstract", (PyCFunction)_wrap_g_type_is_abstract, METH_NOARGS },
+    { "is_value_abstract", (PyCFunction)_wrap_g_type_is_value_abstract, METH_NOARGS },
+    { "is_value_type", (PyCFunction)_wrap_g_type_is_value_type, METH_NOARGS },
+    { "has_value_table", (PyCFunction)_wrap_g_type_has_value_table, METH_NOARGS },
+    { "from_name", (PyCFunction)_wrap_g_type_from_name, METH_VARARGS | METH_STATIC },
+    { "is_a", (PyCFunction)_wrap_g_type_is_a, METH_VARARGS },
+    { NULL,  0, 0 }
+};
+    
+static int
+pyg_type_wrapper_init(PyGTypeWrapper *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "object", NULL };
+    PyObject *py_object;
+    GType type;
+    
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+				     "O:GType.__init__",
+				     kwlist, &py_object))
+        return -1;
+
+    if (!(type = pyg_type_from_object(py_object)))
+	return -1;
+
+    self->type = type;
+
+    return 0;
+}
+
 PyTypeObject PyGTypeWrapper_Type = {
     PyObject_HEAD_INIT(NULL)
     0,
@@ -78,8 +282,26 @@ PyTypeObject PyGTypeWrapper_Type = {
     (hashfunc)pyg_type_wrapper_hash,
     (ternaryfunc)0,
     (reprfunc)0,
-    0L,0L,0L,0L,
-    NULL
+    (getattrofunc)0,		
+    (setattrofunc)0,		
+    0,				
+    Py_TPFLAGS_DEFAULT, 
+    NULL,
+    (traverseproc)0,		
+    (inquiry)0,
+    (richcmpfunc)0,	
+    0,             
+    (getiterfunc)0,
+    (iternextfunc)0,
+    _PyGTypeWrapper_methods,
+    0,
+    _PyGTypeWrapper_getsets,
+    NULL,		
+    NULL,		
+    (descrgetfunc)0,	
+    (descrsetfunc)0,	
+    0,                 
+    (initproc)pyg_type_wrapper_init
 };
 
 /**
@@ -324,7 +546,7 @@ typedef struct {
 } PyGTypeMarshal;
 static GQuark pyg_type_marshal_key = 0;
 
-PyGTypeMarshal *
+static PyGTypeMarshal *
 pyg_type_lookup(GType type)
 {
     GType	ptype = type;
@@ -340,8 +562,8 @@ pyg_type_lookup(GType type)
 }
 
 /**
- * pyg_register_type:
- * @boxed_type: the GType for the new type
+ * pyg_register_gtype_custom:
+ * @gtype: the GType for the new type
  * @from_func: a function to convert GValues to Python objects
  * @to_func: a function to convert Python objects to GValues
  *
@@ -350,40 +572,20 @@ pyg_type_lookup(GType type)
  * handlers.
  */
 
-static void
-pyg_register_gtype(GType type,
-		   fromvaluefunc from_func,
-		   tovaluefunc to_func)
+void
+pyg_register_gtype_custom(GType gtype,
+		                  fromvaluefunc from_func,
+                          tovaluefunc to_func)
 {
     PyGTypeMarshal *tm;
 
     if (!pyg_type_marshal_key)
-	pyg_type_marshal_key = g_quark_from_static_string("PyGType::marshal");
+        pyg_type_marshal_key = g_quark_from_static_string("PyGType::marshal");
 
     tm = g_new(PyGTypeMarshal, 1);
     tm->fromvalue = from_func;
     tm->tovalue = to_func;
-    g_type_set_qdata(type, pyg_type_marshal_key, tm);
-}
-
-/**
- * pyg_register_boxed_custom:
- * @boxed_type: the GType for boxed type
- * @from_func: a function to convert GValues to Python objects
- * @to_func: a function to convert Python objects to GValues
- *
- * The standard way of wrapping boxed types in PyGTK is to create a
- * subclass of gobject.GBoxed and register it with
- * pyg_register_boxed().  In some cases however, it is useful to have
- * fine grained control over how a particular type is represented in
- * Python.  This function allows you to register such a handler.
- */
-void
-pyg_register_boxed_custom(GType boxed_type,
-			  fromvaluefunc from_func,
-			  tovaluefunc to_func)
-{
-    pyg_register_gtype(boxed_type, from_func, to_func);
+    g_type_set_qdata(gtype, pyg_type_marshal_key, tm);
 }
 
 static int
@@ -886,16 +1088,26 @@ pyg_closure_marshal(GClosure *closure,
     }
     ret = PyObject_CallObject(pc->callback, params);
     if (ret == NULL) {
-	PyErr_Print();
+	if (pc->exception_handler)
+	    pc->exception_handler(return_value, n_param_values, param_values);
+	else
+	    PyErr_Print();
 	goto out;
     }
-    if (return_value)
-	pyg_value_from_pyobject(return_value, ret);
+    
+    if (return_value && pyg_value_from_pyobject(return_value, ret) != 0) {
+	PyErr_SetString(PyExc_TypeError,
+			"can't convert return value to desired type");
+	
+	if (pc->exception_handler)
+	    pc->exception_handler(return_value, n_param_values, param_values);
+	else
+	    PyErr_Print();
+    }
     Py_DECREF(ret);
     
  out:
     Py_DECREF(params);
-    
     pyg_gil_state_release(state);
 }
 
@@ -939,6 +1151,28 @@ pyg_closure_new(PyObject *callback, PyObject *extra_args, PyObject *swap_data)
     return closure;
 }
 
+/**
+ * pyg_closure_set_exception_handler:
+ * @closure: a closure created with pyg_closure_new()
+ * @handler: the handler to call when an exception occurs or NULL for none
+ *
+ * Sets the handler to call when an exception occurs during closure invocation.
+ * The handler is responsible for providing a proper return value to the 
+ * closure invocation. If @handler is %NULL, the default handler will be used.
+ * The default handler prints the exception to stderr and doesn't touch the
+ * closure's return value.
+ */
+void
+pyg_closure_set_exception_handler(GClosure *closure,
+				  PyClosureExceptionHandler handler)
+{
+    PyGClosure *pygclosure;
+    
+    g_return_if_fail(closure != NULL);
+
+    pygclosure = (PyGClosure *)closure;
+    pygclosure->exception_handler = handler;
+}
 /* -------------- PySignalClassClosure ----------------- */
 /* a closure used for the `class closure' of a signal.  As this gets
  * all the info from the first argument to the closure and the
